# Copyright (C) 2008-2009 Open Society Institute
#               Thomas Moroz: tmoroz@sorosny.org
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License Version 2 as published
# by the Free Software Foundation.  You may not use, modify or distribute
# this program under any other version of the GNU General Public License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import logging

from pyramid.httpexceptions import HTTPFound
from pyramid.renderers import render_to_response
from pyramid.security import authenticated_userid
from pyramid.security import has_permission
from pyramid.security import remember
from pyramid.url import resource_url
import pyotp

from karl.utils import find_profiles, find_site
from karl.views.api import TemplateAPI
from karl.views.utils import clean_security_code
from karl.views.utils import two_factor_key

log = logging.getLogger(__name__)


def check_otp(two_factor_secret, auth_code=''):
    if two_factor_secret is None:
        return
    totp = pyotp.TOTP(two_factor_secret)
    return totp.verify(clean_security_code(auth_code))


def verify_view(context, request):
    came_from = request.session.get('came_from', request.url)
    site = find_site(context)
    request.session['came_from'] = came_from
    page_title = ''
    api = TemplateAPI(context, request, page_title)
    template_data = {
        'api': api,
        'app_url': request.application_url,
        'secret': '',
        'login_form_url': resource_url(site, request, 'login.html'),
        'outer_class': 'two-factor-verify',
    }
    if request.params.get('form.submitted') is not None:
        # verify code
        userid = request.session.get('partial_auth')
        if userid is None:
            login_form = request.resource_url(context, 'login.html')
            return HTTPFound(location=login_form)
        profiles = find_profiles(context)
        profile = profiles.get(userid)
        auth_code = request.POST.get('auth_code')
        if check_otp(profile.two_factor_secret, auth_code):
            max_age = request.session.pop('max_age')
            del(request.session['partial_auth'])
            remember_headers = remember(request, userid, max_age=max_age)
            env = request.environ
            settings = request.registry.settings
            secret = settings['who_secret']
            tfk = two_factor_key(userid, request, secret)
            cur_domain = env.get(
                'HTTP_HOST',
                env.get('SERVER_NAME')
            ).split(':')[0]
            cookie = ('Set-Cookie', '%s="%s"; Path=/; Domain=%s%s' % (
                'return_2fa', tfk,
                cur_domain, '; Max-Age=2592000',  # 30 days
            ))
            remember_headers.append(cookie)
            came_from = request.session.pop('came_from')
            return HTTPFound(headers=remember_headers, location=came_from)
        else:
            api.status_message = 'Invalid code'
    response = render_to_response(
        'templates/two_factor.pt',
        template_data,
        request=request,
    )
    return response


def setup_view(profile, request):
    login = profile.__name__
    came_from = request.session.get('came_from', request.url)
    request.session['came_from'] = came_from
    page_title = 'Enable Two-Factor Authentication (2FA)'
    api = TemplateAPI(profile, request, page_title)
    template_data = {
        'api': api,
        'nothing': '',
        'app_url': request.application_url,
        'secret': '',
        'skip_url': came_from,
        'outer_class': 'two-factor-setup',
    }

    if request.params.get('form.submitted') is not None:
        # verify code
        auth_code = request.POST.get('auth_code')
        if check_otp(request.session.get('new_secret'), auth_code):
            # keep secret in profile if code was confirmed
            profile.two_factor_secret = request.session.pop('new_secret')
            redirect = request.session.pop('came_from')
        else:
            redirect = request.resource_url(
                profile,
                'two_factor_setup.html',
                query={'reason': 'Invalid code'},
            )
        return HTTPFound(location=redirect)

    if profile.two_factor_secret is None:
        # unconfigured - generate and show code
        if 'new_secret' in request.session:
            new_secret = request.session['new_secret']
        else:
            new_secret = pyotp.random_base32()
            request.session['new_secret'] = new_secret
        totp = pyotp.TOTP(new_secret)
        prov_url = totp.provisioning_uri(
            login,
            issuer_name=api.system_name,
        )
        template_data['prov_url'] = prov_url
        template_data['secret'] = new_secret

    api.status_message = request.params.get('reason')
    layout = request.layout_manager.layout
    layout.page_title = page_title
    response = render_to_response(
        'templates/two_factor.pt',
        template_data,
        request=request,
    )
    return response


def reset_view(profile, request):
    login = profile.__name__
    page_title = 'Reset Two-Factor Authentication'
    api = TemplateAPI(profile, request, page_title)
    profile_url = request.resource_url(profile)
    template_data = {
        'api': api,
        'profile_url': profile_url,
        'admin_reset': 0,
    }
    if (authenticated_userid(request) == login):
        redirect = request.resource_url(
            profile,
            'two_factor_setup.html',
        )
        request.session['came_from'] = profile_url
        if profile.two_factor_secret is None:
            # 2FA was never configured; go to setup
            return HTTPFound(location=redirect)
        if request.params.get('form.submitted') is not None:
            # verify code
            auth_code = request.POST.get('auth_code')
            if check_otp(profile.two_factor_secret, auth_code):
                # reset and go back to setup
                profile.two_factor_secret = None
                return HTTPFound(location=redirect)
            else:
                api.status_message = 'Invalid code'
    elif has_permission('administer', profile, request):
        profile.two_factor_secret = None
        api.status_message = 'Two-factor authentication reset for user'
        template_data['admin_reset'] = 1
    response = render_to_response(
        'templates/two_factor_reset.pt',
        template_data,
        request=request,
    )
    return response
